#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar  1 17:47:48 2018


Compute oifits model files from input oifits data files and one input image 

python oifitsmodeler.py test_input_oifitsmodeler.txt


@author: domicia
"""

import numpy as np
from astropy.io import fits
import os
#from subprocess import call
import subprocess
import shlex
from sys import argv
from sys import exit as saida
#import matplotlib.pyplot as plt
#from scipy.ndimage import rotate as imrot

#%%
def findfiles(directory):
    files = []
    for i in os.listdir(directory):  # find all objects in a dir and check them...
        if isFile(directory + i):  # ... is a file.
            if i.endswith(".fits"):
                files.append(i)  # if yes, append it.
    return files



def isFile(object):
    try:
        os.listdir(object)  # tries to get the objects inside of this object
        return False  # if it worked, it's a folder
    except Exception:  # if not, it's a file
        return True

#%%
def compute_chi2(oifits_data_dir, oifits_model_dir, data4chi2, ignore_flagged=1):
    """
    Compute chi^2 from oidata and oimodel files and append into oimodel header
    """
    
    dic={'VIS2' : {'ext':'OI_VIS2', 'data': 'VIS2DATA', 'error':'VIS2ERR' , 'ignore_flagged': ignore_flagged}, \
         'T3AMP': {'ext':'OI_T3'  , 'data': 'T3AMP'   , 'error':'T3AMPERR', 'ignore_flagged': ignore_flagged}, \
         'T3PHI': {'ext':'OI_T3'  , 'data': 'T3PHI'   , 'error':'T3PHIERR', 'ignore_flagged': ignore_flagged} }

    #Flag to check if chi^2 calculations worked well
    #flag=True

    #List of observed OIFITS
    all_oifiles=findfiles(oifits_data_dir)

    #List of modeled OIFITS
    all_oimodels=findfiles(oifits_model_dir)
    #print(all_oimodels)

    #Compute chi^2 for each couple of data and model oifits files
    for oifiles_i in all_oifiles:
        oidata_name=oifiles_i.split('.fits')[0]
        print('{}'.format(oidata_name))
        #print([s for s in all_oimodels if oidata_name in s])
        #PUT A TEST HERE TO CHECK THAT THERE IS ONLY ONE FILE IN THIS LIST
        oimodels_i=[s for s in all_oimodels if oidata_name in s][0]
        print('{}'.format(oimodels_i))
        

        #Open data and model oifits files
        hdldata  = fits.open(oifits_data_dir+oifiles_i)
        #Opening in update mode so that header can be changed
        hdlmodel = fits.open(oifits_model_dir+oimodels_i, mode='update') 

        #Update header
        hdlmodel['Primary'].header['comment'] = 'chi^2 for this modeled OIFITS file'
        
        #Initialize total chi^2 variables
        ndata_tot=0        
        chi2_tot=0.0


        for i in data4chi2:
            print('Treating data type {}'.format(i)) 
            #Create new variables for an easier visuzalisation
            obs=hdldata[dic[i]['ext']].data.field(dic[i]['data'])
            err=hdldata[dic[i]['ext']].data.field(dic[i]['error'])
            mod=hdlmodel[dic[i]['ext']].data.field(dic[i]['data'])
            chi2=(((obs-mod)/err)**2.).flatten()
            
            #Use flags in oidata to remove flagged data from chi^2 if required
            #Add comment in header about this
            print(hdldata[dic[i]['ext']].data.FLAG.shape)
            print(obs.shape)
            if dic[i]['ignore_flagged'] == 1:
                hdlmodel['Primary'].header['comment'] = 'flagged data ignored in chi^2 '+dic[i]['data']
                flags=hdldata[dic[i]['ext']].data.FLAG.flatten()
                #idx_not_flagged=np.where(flags == 70)
                chi2=chi2[np.where(flags == 70)]
                
            #Compute chi^2 info for the present data type and update header
            ndata=chi2.size
            chi2=np.sum(chi2)
            print('{}: chi^2 = {}; ndata = {}'.format(i, chi2, ndata)) 
            hdlmodel['Primary'].header.set('X2'+i, chi2, i+' chi^2')
            hdlmodel['Primary'].header.set('N'+i, ndata, 'nb of '+i+' data')
            
            #Update total chi^2 variables
            ndata_tot=ndata_tot+ndata            
            chi2_tot=chi2_tot+chi2      
            
            """
            Alternative way to compute chi^2 without removing flagged data
            chi2_test=np.sum(((obs-mod)/err)**2.)
            #print(obs.size)
            #print(len(obs))
            ndata_test=obs.size  #different from len(obs)
            print('Tests: chi^2 = {}; ndata = {}'.format(chi2_test, ndata_test))
            """

        
        hdlmodel['Primary'].header.set('CHI2TOT', chi2_tot, 'total chi^2')
        hdlmodel['Primary'].header.set('NTOT', ndata_tot, 'total nb of data')
            
        hdlmodel.flush()  # changes are written back to original.fits
        # closing the file will also flush any changes and prevent further writing
        hdldata.close()
        hdlmodel.close()


    return
#%%    


"""
Useful websites

#https://stackoverflow.com/questions/3777301/how-to-call-a-shell-script-from-python-code
#base_path='/Users/domicia/Desktop/JMMC_AMHRA_outil_model_oifits/'
#subprocess.call(shlex.split(base_path+'test.sh param1 param2'))
#subprocess.call(shlex.split(base_path+'./test.sh /Users/domicia/Desktop/JMMC_AMHRA_outil_model_oifits/ image_disco_amhra.fits Achernar_all_Pionier_oiDataCalib_2012-09-17.fits'))

See:
http://docs.astropy.org/en/stable/io/fits/
https://docs.scipy.org/doc/numpy/reference/generated/numpy.recarray.html#numpy.recarray
https://stackoverflow.com/questions/33711985/flattening-a-list-of-numpy-arrays
https://docs.scipy.org/doc/numpy-1.14.0/reference/generated/numpy.ndarray.flatten.html
https://docs.scipy.org/doc/numpy-1.14.0/reference/generated/numpy.ndarray.size.html
http://www.astropy.org/astropy-tutorials/FITS-images.html
https://docs.scipy.org/doc/scipy-0.16.0/reference/generated/scipy.ndimage.interpolation.rotate.html
https://stackoverflow.com/questions/46657423/rotated-image-coordinates-after-scipy-ndimage-interpolation-rotate

#The data structure used to represent FITS tables is called a FITS_rec and is derived from the numpy.recarray interface. 
#When creating a new table HDU the individual column arrays will be assembled into a single FITS_rec array.
#https://docs.scipy.org/doc/numpy/reference/generated/numpy.recarray.html#numpy.recarray
#Record arrays allow the fields to be accessed as members of the array, using arr.x and arr.y.


Examples of how to recover data and info from the oifits file:
hdulist.info()
hdulist[0].header
hdulist[1].header['ARRNAME'] or hdulist['OI_ARRAY'].header['ARRNAME']
len(hdulist)
hdulist[4].data.VIS2DATA  or hdulist['OI_VIS2'].data.VIS2DATA
hdulist['OI_VIS2'].data.VIS2DATA.shape
hdulist['OI_ARRAY'].header.set('observer', 'Edwin Hubble')  #update header
hdulist.flush()  # changes are written back to original.fits
hdldata['OI_VIS2'].data.field('VIS2DATA')[0,:] or hdldata['OI_VIS2'].data.field(4)[0,:] or hdldata['OI_VIS2'].data.VIS2DATA[0,:]
type(hdldata['OI_VIS2'].data.field('VIS2DATA')) => numpy.ndarray


"""



#Check if input is ok
if len(argv) == 1:
    saida("Please give input file with 'base_dir' 'oimodel' 'data4chi2' 'ignore_flagged'.")
elif len(argv) == 2:
    input_file = argv[1]



#Read input file into pars dictionary
pars = {}
with open(input_file, 'r') as f: 
    for line in f:
        if not (line.startswith('#') or line.startswith('*') or not line.strip()):
            pars[line.split('#')[0].strip().split('=')[0].split()[0]] = \
                  (line.split('#')[0].strip().split('=')[1].split()[0]) 




#Copy pars values into more explicit variables and convert some quantities
base_dir =pars['base_dir']
oimodel  =pars['oimodel']
data4chi2=pars['data4chi2'].split('+')
ignore_flagged=int(pars['ignore_flagged'])


#Base directory
#base_dir='/Users/domicia/Desktop/JMMC_AMHRA_outil_model_oifits/'

#OIFITS data directory
oifits_data_dir=base_dir+'oifits_data/'

#Directory for the output OIFITS model files
oifits_model_dir=base_dir+'oifits_model/'

#Input model image
oimodel_dir=base_dir
#oimodel='image_disco_amhra_UD6mas.fits'
#oimodel='image_disco_amhra.fits'



#Info for chi^2 calculations
#Observables to be considered for the chi^2
#Same syntax in the oifits files
#Possible values/ 'VIS2','T3AMP','T3PHI', 'VIS', 'VISPHI'
#data4chi2=['VIS2','T3AMP','T3PHI']

#Flag to ignore flagged data (1) or to keep all data (0)
#ignore_flagged=1







"""
Rotate image ?

see:
https://stackoverflow.com/questions/46657423/rotated-image-coordinates-after-scipy-ndimage-interpolation-rotate
https://docs.scipy.org/doc/scipy-0.16.0/reference/generated/scipy.ndimage.interpolation.rotate.html

hdlimg = fits.open(oimodel_dir+oimodel) 
hdlimg.info()
hdlimg[0].header
hdlimg[0].data.shape

ilam=0
img=hdlimg[0].data[ilam,:,:]
plt.imshow(img, cmap='gray')

"""



#List of observed OIFITS
all_oifiles=findfiles(oifits_data_dir)

for oifiles_i in all_oifiles:
    print('Treating file {}'.format(oifiles_i))
    oimodel_output=oifits_model_dir+oifiles_i.split('.fits')[0]+'_'+oimodel
    subprocess.call(shlex.split(base_dir+'./model_oifits_from_aspro2.sh'+' '+\
        base_dir+' '+base_dir+oimodel+' '+oifits_data_dir+oifiles_i+' '+oimodel_output))
   

compute_chi2(oifits_data_dir, oifits_model_dir, data4chi2)




#%%

